@@ -21,6 +21,7 @@ const Clutter = imports.gi.Clutter;
 const St = imports.gi.St;
 const Gio = imports.gi.Gio;
 const GLib = imports.gi.GLib;
+const GObject = imports.gi.GObject;
 const Shell = imports.gi.Shell;
 
 const Signals = imports.signals;
@@ -62,11 +63,10 @@ var StoredCoordinates = {
     ASSIGN:2,
 };
 
-class Placeholder extends St.Bin {
-    constructor() {
-        super();
-    }
-}
+var Placeholder = GObject.registerClass({
+    GTypeName: 'DesktopIcons_Placeholder',
+}, class Placeholder extends St.Bin {
+});
 
 var DesktopGrid = class {
 
@@ -92,6 +92,11 @@ var DesktopGrid = class {
         });
         this.actor._delegate = this;
 
+        this.actor.connect('style-changed', () => {
+            if (this.isFirst)
+                Extension.desktopManager.scheduleReLayoutChildren();
+        });
+
         this._grid = new St.Widget({
             name: 'DesktopGrid',
             layout_manager: this.layout,
@@ -103,8 +108,7 @@ var DesktopGrid = class {
         });
         this.actor.add_child(this._grid);
 
-        this._renamePopup = new RenamePopup(this);
-        this.actor.add_child(this._renamePopup.actor);
+        this._menuManager = new PopupMenu.PopupMenuManager(this.actor);
 
         this._bgManager._container.add_child(this.actor);
 
@@ -127,6 +131,22 @@ var DesktopGrid = class {
         this._grid.connect('key-press-event', this._onKeyPress.bind(this));
 
         this._grid.connect('allocation-changed', () => Extension.desktopManager.scheduleReLayoutChildren());
+
+        let themeNodeBase = this.actor.get_theme_node();
+        let themeContext = St.ThemeContext.get_for_stage(global.stage);
+        let themeNode = St.ThemeNode.new(themeContext, null,
+            themeNodeBase.get_theme(), themeNodeBase.get_element_type(),
+            null, "file-item", null, "fake-id {}");
+        this._extra_width =  themeNode.get_margin(St.Side.LEFT) +
+                             themeNode.get_margin(St.Side.RIGHT) +
+                             themeNode.get_border_width(St.Side.LEFT) +
+                             themeNode.get_border_width(St.Side.RIGHT) +
+                             themeNode.get_horizontal_padding();
+        this._extra_height = themeNode.get_margin(St.Side.TOP) +
+                             themeNode.get_margin(St.Side.BOTTOM) +
+                             themeNode.get_border_width(St.Side.TOP) +
+                             themeNode.get_border_width(St.Side.BOTTOM) +
+                             themeNode.get_vertical_padding();
     }
 
     _onKeyPress(actor, event) {
@@ -190,6 +210,7 @@ var DesktopGrid = class {
             this._bgManager.backgroundActor.disconnect(this._bgDestroyedId);
         this._bgDestroyedId = 0;
         this._bgManager = null;
+        this._menuManager = null;
     }
 
     _onNewFolderClicked() {
@@ -308,8 +329,12 @@ var DesktopGrid = class {
     }
 
     _syncUndoRedo() {
-        this._undoMenuItem.actor.visible = DBusUtils.NautilusFileOperationsProxy.UndoStatus == UndoStatus.UNDO;
-        this._redoMenuItem.actor.visible = DBusUtils.NautilusFileOperationsProxy.UndoStatus == UndoStatus.REDO;
+        let useItemActors = !ExtensionUtils.versionCheck(['3.28', '3.30', '3.32'], Config.PACKAGE_VERSION);
+        let undoActor = useItemActors ? this._undoMenuItem : this._undoMenuItem.actor;
+        let redoActor = useItemActors ? this._redoMenuItem : this._redoMenuItem.actor;
+
+        undoActor.visible = DBusUtils.NautilusFileOperationsProxy.UndoStatus == UndoStatus.UNDO;
+        redoActor.visible = DBusUtils.NautilusFileOperationsProxy.UndoStatus == UndoStatus.REDO;
     }
 
     _undoStatusChanged(proxy, properties, test) {
@@ -368,7 +393,7 @@ var DesktopGrid = class {
         /* Since the handler is in the press event it needs to ignore the release event
          * to not immediately close the menu on release
          */
-        this.actor._desktopBackgroundManager.ignoreRelease();
+        this._menuManager.ignoreRelease();
     }
 
     _addFileItemTo(fileItem, column, row, coordinatesAction) {
@@ -376,7 +401,7 @@ var DesktopGrid = class {
         placeholder.child = fileItem.actor;
         this._fileItems.push(fileItem);
         let selectedId = fileItem.connect('selected', this._onFileItemSelected.bind(this));
-        let renameId = fileItem.connect('rename-clicked', this.doRename.bind(this));
+        let renameId = fileItem.connect('rename-clicked', this._onRenameClicked.bind(this));
         this._fileItemHandlers.set(fileItem, [selectedId, renameId]);
 
         /* If this file is new in the Desktop and hasn't yet
@@ -393,6 +418,7 @@ var DesktopGrid = class {
 
     addFileItemCloseTo(fileItem, x, y, coordinatesAction) {
         let [column, row] = this._getEmptyPlaceClosestTo(x, y, coordinatesAction);
+        fileItem.set_margins(this._extra_width, this._extra_height);
         this._addFileItemTo(fileItem, column, row, coordinatesAction);
     }
 
@@ -508,24 +534,26 @@ var DesktopGrid = class {
 
     _addDesktopBackgroundMenu() {
         this.actor._desktopBackgroundMenu = this._createDesktopBackgroundMenu();
-        this.actor._desktopBackgroundManager = new PopupMenu.PopupMenuManager({ actor: this.actor });
-        this.actor._desktopBackgroundManager.addMenu(this.actor._desktopBackgroundMenu);
+        this._menuManager.addMenu(this.actor._desktopBackgroundMenu);
 
         this.actor.connect('destroy', () => {
             this.actor._desktopBackgroundMenu.destroy();
             this.actor._desktopBackgroundMenu = null;
-            this.actor._desktopBackgroundManager = null;
         });
     }
 
     _getMaxColumns() {
         let gridWidth = this._grid.allocation.x2 - this._grid.allocation.x1;
-        return Math.floor(gridWidth / Prefs.get_desired_width(St.ThemeContext.get_for_stage(global.stage).scale_factor));
+        let scaleFactor = St.ThemeContext.get_for_stage(global.stage).scale_factor;
+        let desiredWidth = Prefs.getDesiredWidth(scaleFactor, this._extra_width);
+        return Math.floor(gridWidth / desiredWidth);
     }
 
     _getMaxRows() {
         let gridHeight = this._grid.allocation.y2 - this._grid.allocation.y1;
-        return Math.floor(gridHeight / Prefs.get_desired_height(St.ThemeContext.get_for_stage(global.stage).scale_factor));
+        let scaleFactor = St.ThemeContext.get_for_stage(global.stage).scale_factor;
+        let desiredHeight = Prefs.getDesiredHeight(scaleFactor, this._extra_height);
+        return Math.floor(gridHeight / desiredHeight);
     }
 
     acceptDrop(source, actor, x, y, time) {
@@ -568,125 +596,142 @@ var DesktopGrid = class {
         this._grid.grab_key_focus();
     }
 
+    _onRenameClicked(fileItem) {
+        if (fileItem.menu && fileItem.menu.isOpen) {
+            let id = fileItem.menu.connect('menu-closed', () => {
+                fileItem.menu.disconnect(id);
+                this.doRename(fileItem);
+            });
+        } else {
+            this.doRename(fileItem);
+        }
+    }
+
     doRename(fileItem) {
-        this._renamePopup.onFileItemRenameClicked(fileItem);
+        let renamePopup = new RenamePopup(fileItem);
+        this._menuManager.addMenu(renamePopup);
+        this._menuManager.ignoreRelease();
+
+        renamePopup.connect('menu-closed', () => renamePopup.destroy());
+        renamePopup.open();
     }
 };
 
-var RenamePopup = class {
+var RenamePopupMenuItem = GObject.registerClass(
+class RenamePopupMenuItem extends PopupMenu.PopupBaseMenuItem {
+    _init(fileItem) {
+        super._init({
+            style_class: 'rename-popup-item',
+            reactive: false,
+        });
 
-    constructor(grid) {
-        this._source = null;
-        this._isOpen = false;
+        if (PopupMenu.Ornament.HIDDEN !== undefined)
+            this.setOrnament(PopupMenu.Ornament.HIDDEN);
+        else /* Support version prior 3.34.1 */
+            this._ornamentLabel.visible = false;
 
-        this._renameEntry = new St.Entry({ hint_text: _("Enter file name…"),
-                                           can_focus: true,
-                                           x_expand: true });
-        this._renameEntry.clutter_text.connect('activate', this._onRenameAccepted.bind(this));
-        this._renameOkButton= new St.Button({ label: _("OK"),
-                                              style_class: 'app-view-control button',
-                                              button_mask: St.ButtonMask.ONE | St.ButtonMask.THREE,
-                                              reactive: true,
-                                              can_focus: true,
-                                              x_expand: true });
-        this._renameCancelButton = new St.Button({ label: _("Cancel"),
-                                                   style_class: 'app-view-control button',
-                                                   button_mask: St.ButtonMask.ONE | St.ButtonMask.THREE,
-                                                   reactive: true,
-                                                   can_focus: true,
-                                                   x_expand: true });
-        this._renameCancelButton.connect('clicked', () => { this._onRenameCanceled(); });
-        this._renameOkButton.connect('clicked', () => { this._onRenameAccepted(); });
-        let renameButtonsBoxLayout = new Clutter.BoxLayout({ homogeneous: true });
-        let renameButtonsBox = new St.Widget({ layout_manager: renameButtonsBoxLayout,
-                                               x_expand: true });
-        renameButtonsBox.add_child(this._renameCancelButton);
-        renameButtonsBox.add_child(this._renameOkButton);
+        this._fileItem = fileItem;
 
-        let renameContentLayout = new Clutter.BoxLayout({ spacing: 6,
-                                                          orientation: Clutter.Orientation.VERTICAL });
-        let renameContent = new St.Widget({ style_class: 'rename-popup',
-                                            layout_manager: renameContentLayout,
-                                            x_expand: true });
-        renameContent.add_child(this._renameEntry);
-        renameContent.add_child(renameButtonsBox);
+        // Entry
+        this._entry = new St.Entry({
+            hint_text: _("Enter file name…"),
+            x_expand: true,
+            width: 200,
+        });
+        this.add_child(this._entry);
 
-        this._boxPointer = new BoxPointer.BoxPointer(St.Side.TOP, { can_focus: false, x_expand: false });
-        this.actor = this._boxPointer.actor;
-        this.actor.style_class = 'popup-menu-boxpointer';
-        this.actor.add_style_class_name('popup-menu');
-        this.actor.visible = false;
-        this._boxPointer.bin.set_child(renameContent);
+        this._entry.clutter_text.connect(
+            'notify::text', this._validate.bind(this));
+        this._entry.clutter_text.connect(
+            'activate', this._onRenameAccepted.bind(this));
 
-        this._grabHelper = new GrabHelper.GrabHelper(grid.actor, { actionMode: Shell.ActionMode.POPUP });
-        this._grabHelper.addActor(this.actor);
+        // Rename button
+        this._button = new St.Button({
+            style_class: 'button',
+            reactive: true,
+            button_mask: St.ButtonMask.ONE | St.ButtonMask.TWO,
+            can_focus: true,
+            label: _('Rename'),
+        });
+        this.add_child(this._button);
+        this._button.connect('clicked', this._onRenameAccepted.bind(this));
     }
 
-    _popup() {
-        if (this._isOpen)
-            return;
-
-        this._isOpen = this._grabHelper.grab({ actor: this.actor,
-                                               onUngrab: this._popdown.bind(this) });
-
-        if (!this._isOpen) {
-            this._grabHelper.ungrab({ actor: this.actor });
-            return;
-        }
-
-        this._boxPointer.setPosition(this._source.actor, 0.5);
-        if (ExtensionUtils.versionCheck(['3.28', '3.30'], Config.PACKAGE_VERSION))
-            this._boxPointer.show(BoxPointer.PopupAnimation.FADE |
-                                  BoxPointer.PopupAnimation.SLIDE);
-        else
-            this._boxPointer.open(BoxPointer.PopupAnimation.FADE |
-                                  BoxPointer.PopupAnimation.SLIDE);
-
-        this.emit('open-state-changed', true);
+    vfunc_map() {
+        this._entry.text = this._fileItem.displayName;
+        this._entry.clutter_text.set_selection(0, -1);
+        super.vfunc_map();
     }
 
-    _popdown() {
-        if (!this._isOpen)
-            return;
-
-        this._grabHelper.ungrab({ actor: this.actor });
-
-        if (ExtensionUtils.versionCheck(['3.28', '3.30'], Config.PACKAGE_VERSION))
-            this._boxPointer.hide(BoxPointer.PopupAnimation.FADE |
-                                   BoxPointer.PopupAnimation.SLIDE);
-        else
-            this._boxPointer.close(BoxPointer.PopupAnimation.FADE |
-                                  BoxPointer.PopupAnimation.SLIDE);
-
-        this._isOpen = false;
-        this.emit('open-state-changed', false);
+    vfunc_key_focus_in() {
+        super.vfunc_key_focus_in();
+        this._entry.clutter_text.grab_key_focus();
     }
 
-    onFileItemRenameClicked(fileItem) {
-        this._source = fileItem;
+    _isValidFolderName() {
+        let newName = this._entry.text.trim();
 
-        this._renameEntry.text = fileItem.displayName;
+        return newName.length > 0 && newName.indexOf('/') === -1 &&
+               newName != this._fileItem.displayName;
+    }
 
-        this._popup();
-        this._renameEntry.grab_key_focus();
-        this._renameEntry.navigate_focus(null, Gtk.DirectionType.TAB_FORWARD, false);
-        let extensionOffset = DesktopIconsUtil.getFileExtensionOffset(fileItem.displayName, fileItem.isDirectory);
-        this._renameEntry.clutter_text.set_selection(0, extensionOffset);
+    _validate() {
+        this._button.reactive = this._isValidFolderName();
     }
 
     _onRenameAccepted() {
-        this._popdown();
-        DBusUtils.NautilusFileOperationsProxy.RenameFileRemote(this._source.file.get_uri(),
-                                                               this._renameEntry.get_text(),
+        if (!this._isValidFolderName())
+            return;
+
+        DBusUtils.NautilusFileOperationsProxy.RenameFileRemote(
+            this._fileItem.file.get_uri(),
+            this._entry.text.trim(),
             (result, error) => {
                 if (error)
                     throw new Error('Error renaming file: ' + error.message);
             }
         );
-    }
 
-    _onRenameCanceled() {
-        this._popdown();
+        this.activate(Clutter.get_current_event());
+    }
+});
+
+var RenamePopup = class RenameFolderMenu extends PopupMenu.PopupMenu {
+    constructor(fileItem) {
+        super(fileItem.actor, 0.5, St.Side.TOP);
+        this.actor.add_style_class_name('rename-popup');
+
+        // We want to keep the item hovered while the menu is up
+        this.blockSourceEvents = true;
+
+        let menuItem = new RenamePopupMenuItem(fileItem);
+        this.addMenuItem(menuItem);
+
+        if (this.focusActor !== undefined) {
+            // Focus the text entry on menu pop-up, works starting 3.34.1
+            this.focusActor = menuItem;
+        } else {
+            this.connect('open-state-changed', (_menu, state) => {
+                if (state)
+                    this._openId = GLib.idle_add(GLib.PRIORITY_DEFAULT, () => {
+                        delete this._openId;
+                        menuItem.grab_key_focus()
+                    });
+                else if (this._openId)
+                    GLib.source_remove(this._openId);
+            });
+        }
+
+        // Chain our visibility and lifecycle to that of the fileItem source
+        this._fileItemMappedId = fileItem.actor.connect('notify::mapped', () => {
+            if (!fileItem.actor.mapped)
+                this.close();
+        });
+        fileItem.actor.connect('destroy', () => {
+            fileItem.actor.disconnect(this._fileItemMappedId);
+            this.destroy();
+        });
+
+        Main.uiGroup.add_actor(this.actor);
     }
 };
-Signals.addSignalMethods(RenamePopup.prototype);
