@@ -183,6 +183,7 @@ var DesktopManager = GObject.registerClass({
     }
 
     _addDesktopIcons() {
+        let first_element = true;
         forEachBackgroundManager(bgManager => {
             let newGrid = new DesktopGrid.DesktopGrid(bgManager);
             newGrid.actor.connect('destroy', (actor) => {
@@ -193,6 +194,8 @@ var DesktopManager = GObject.registerClass({
                         break;
                     }
             });
+            newGrid.isFirst = first_element;
+            first_element = false;
             this._desktopGrids[bgManager._monitorIndex] = newGrid;
         });
 
@@ -204,7 +207,7 @@ var DesktopManager = GObject.registerClass({
         this._desktopGrids = {};
     }
 
-    /** 
+    /**
      * Initialize rubberband color from the GTK rubberband class
      * */
     _initRubberBandColor() {
@@ -342,8 +345,10 @@ var DesktopManager = GObject.registerClass({
             case ATTRIBUTE_CHANGED:
                 /* a file changed, rather than the desktop itself */
                 let desktopDir = DesktopIconsUtil.getDesktopDir();
-                if (file.get_uri() != desktopDir.get_uri())
+                if (file.get_uri() != desktopDir.get_uri()) {
+                    fileItem.onAttributeChanged();
                     return;
+                }
 
                 if (this._queryFileInfoCancellable)
                     this._queryFileInfoCancellable.cancel();
