@@ -30,6 +30,7 @@ const GnomeDesktop = imports.gi.GnomeDesktop;
 const Mainloop = imports.mainloop;
 const Signals = imports.signals;
 
+const Config = imports.misc.config;
 const Background = imports.ui.background;
 const Main = imports.ui.main;
 const PopupMenu = imports.ui.popupMenu;
@@ -78,8 +79,6 @@ var FileItem = class {
         this.actor = new St.Bin({ visible: true });
         this.actor.set_fill(true, true);
         let scaleFactor = St.ThemeContext.get_for_stage(global.stage).scale_factor;
-        this.actor.set_height(Prefs.get_desired_height(scaleFactor));
-        this.actor.set_width(Prefs.get_desired_width(scaleFactor));
         this.actor._delegate = this;
         this.actor.connect('destroy', () => this._onDestroy());
 
@@ -118,30 +117,18 @@ var FileItem = class {
         /* Set the metadata and update relevant UI */
         this._updateMetadataFromFileInfo(fileInfo);
 
-        if (this._isDesktopFile) {
-            /* watch for the executable bit being removed or added */
-            this._monitorDesktopFile = this._file.monitor_file(Gio.FileMonitorFlags.NONE, null);
-            this._monitorDesktopFileId = this._monitorDesktopFile.connect('changed',
-                (obj, file, otherFile, eventType) => {
-                    switch(eventType) {
-                        case Gio.FileMonitorEvent.ATTRIBUTE_CHANGED:
-                            this._refreshMetadataAsync(true);
-                            break;
-                    }
-                });
-        }
-
-        this._createMenu();
+        this._menuManager = null;
+        this._menu = null;
         this._updateIcon();
 
         this._isSelected = false;
         this._primaryButtonPressed = false;
-        if (this._attributeCanExecute && !this._isDesktopFile)
+        if (this._attributeCanExecute && !this._isValidDesktopFile)
             this._execLine = this.file.get_path();
         if (fileExtra == Prefs.FileType.USER_DIRECTORY_TRASH) {
             // if this icon is the trash, monitor the state of the directory to update the icon
             this._trashChanged = false;
-            this._trashInitializeCancellable = null;
+            this._queryTrashInfoCancellable = null;
             this._scheduleTrashRefreshId = 0;
             this._monitorTrashDir = this._file.monitor_directory(Gio.FileMonitorFlags.WATCH_MOVES, null);
             this._monitorTrashId = this._monitorTrashDir.connect('changed', (obj, file, otherFile, eventType) => {
@@ -162,13 +149,25 @@ var FileItem = class {
             });
         }
 
-        Extension.desktopManager.connect('notify::writable-by-others', () => {
-            if (!this._isDesktopFile)
+        this._writebleByOthersId = Extension.desktopManager.connect('notify::writable-by-others', () => {
+            if (!this._isValidDesktopFile)
                 return;
             this._refreshMetadataAsync(true);
         });
     }
 
+    set_margins(width, height) {
+        let scaleFactor = St.ThemeContext.get_for_stage(global.stage).scale_factor;
+        this.actor.set_width(Prefs.getDesiredWidth(scaleFactor, width));
+        this.actor.set_height(Prefs.getDesiredHeight(scaleFactor, height));
+    }
+
+    onAttributeChanged() {
+        if (this._isDesktopFile) {
+            this._refreshMetadataAsync(true);
+        }
+    }
+
     _onDestroy() {
         /* Regular file data */
         if (this._setMetadataCancellable)
@@ -176,6 +175,8 @@ var FileItem = class {
         if (this._queryFileInfoCancellable)
             this._queryFileInfoCancellable.cancel();
 
+        Extension.desktopManager.disconnect(this._writebleByOthersId);
+
         /* Thumbnailing */
         if (this._thumbnailScriptWatch)
             GLib.source_remove(this._thumbnailScriptWatch);
@@ -197,6 +198,9 @@ var FileItem = class {
             this._queryTrashInfoCancellable.cancel();
         if (this._scheduleTrashRefreshId)
             GLib.source_remove(this._scheduleTrashRefreshId);
+
+        /* Menu */
+        this._removeMenu();
     }
 
     _refreshMetadataAsync(rebuild) {
@@ -213,7 +217,7 @@ var FileItem = class {
                     this._queryFileInfoCancellable = null;
                     this._updateMetadataFromFileInfo(newFileInfo);
                     if (rebuild) {
-                        this._createMenu();
+                        this._recreateMenu();
                         this._updateIcon();
                     }
                 } catch(error) {
@@ -243,8 +247,12 @@ var FileItem = class {
             this._desktopFile = Gio.DesktopAppInfo.new_from_filename(this._file.get_path());
             if (!this._desktopFile) {
                 log(`Couldn’t parse ${this._displayName} as a desktop file, will treat it as a regular file.`);
-                this._isDesktopFile = false;
+                this._isValidDesktopFile = false;
+            } else {
+                this._isValidDesktopFile = true;
             }
+        } else {
+            this._isValidDesktopFile = false;
         }
 
         if (this.displayName != oldLabelText) {
@@ -296,6 +304,7 @@ var FileItem = class {
                     this._thumbnailScriptWatch = GLib.child_watch_add(GLib.PRIORITY_DEFAULT,
                                                                       pid,
                         (pid, exitCode) => {
+                            this._thumbnailScriptWatch = 0;
                             if (exitCode == 0)
                                 this._updateIcon();
                             else
@@ -313,6 +322,7 @@ var FileItem = class {
                 thumbnailFile.load_bytes_async(this._loadThumbnailDataCancellable,
                     (source, result) => {
                         try {
+                            this._loadThumbnailDataCancellable = null;
                             let [thumbnailData, etag_out] = source.load_bytes_finish(result);
                             let thumbnailStream = Gio.MemoryInputStream.new_from_bytes(thumbnailData);
                             let thumbnailPixbuf = GdkPixbuf.Pixbuf.new_from_stream(thumbnailStream, null);
@@ -328,7 +338,7 @@ var FileItem = class {
                                 );
                                 let icon = new Clutter.Actor();
                                 icon.set_content(thumbnailImage);
-                                let width = Prefs.get_desired_width(scaleFactor);
+                                let width = Prefs.getDesiredWidth(scaleFactor, 0);
                                 let height = Prefs.get_icon_size() * scaleFactor;
                                 let aspectRatio = thumbnailPixbuf.width / thumbnailPixbuf.height;
                                 if ((width / height) > aspectRatio)
@@ -435,7 +445,7 @@ var FileItem = class {
             return;
         }
 
-        if (this._attributeCanExecute && !this._isDirectory && !this._isDesktopFile) {
+        if (this._attributeCanExecute && !this._isDirectory && !this._isValidDesktopFile) {
             if (this._execLine)
                 Util.spawnCommandLine(this._execLine);
             return;
@@ -565,8 +575,29 @@ var FileItem = class {
                'border-color: ' + border_color + ';'
     }
 
-    _createMenu() {
-        this._menuManager = new PopupMenu.PopupMenuManager({ actor: this.actor });
+    get menu() {
+        return this._menu;
+    }
+
+    _removeMenu() {
+        if (this._menu != null) {
+            if (this._menuManager != null)
+                this._menuManager.removeMenu(this._menu);
+
+            Main.layoutManager.uiGroup.remove_child(this._menu.actor);
+            this._menu.destroy();
+            this._menu = null;
+        }
+
+        this._menuManager = null;
+    }
+
+    _recreateMenu() {
+        this._removeMenu();
+        let popupManagerArgs = ExtensionUtils.versionCheck(['3.28', '3.30', '3.32'], Config.PACKAGE_VERSION) ?
+            { actor: this.actor } : this.actor;
+
+        this._menuManager = new PopupMenu.PopupMenuManager(popupManagerArgs);
         let side = St.Side.LEFT;
         if (Clutter.get_default_text_direction() == Clutter.TextDirection.RTL)
             side = St.Side.RIGHT;
@@ -584,7 +615,7 @@ var FileItem = class {
             if (this.canRename())
                 this._menu.addAction(_('Rename…'), () => this.doRename());
             this._actionTrash = this._menu.addAction(_('Move to Trash'), () => this._onMoveToTrashClicked());
-            if (this._isDesktopFile && !Extension.desktopManager.writableByOthers && !this._writableByOthers) {
+            if (this._isValidDesktopFile && !Extension.desktopManager.writableByOthers && !this._writableByOthers) {
                 this._menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
                 this._allowLaunchingMenuItem = this._menu.addAction(this._allowLaunchingText,
                                                                     () => this._onAllowDisallowLaunchingClicked());
@@ -611,6 +642,13 @@ var FileItem = class {
         this._menu.actor.hide();
     }
 
+    _ensureMenu() {
+        if (this._menu == null)
+            this._recreateMenu();
+
+        return this._menu;
+    }
+
     _onOpenTerminalClicked () {
         DesktopIconsUtil.launchTerminal(this.file.get_path());
     }
@@ -620,7 +658,7 @@ var FileItem = class {
         if (button == 3) {
             if (!this.isSelected)
                 this.emit('selected', false, false, true);
-            this._menu.toggle();
+            this._ensureMenu().toggle();
             if (this._actionOpenWith) {
                 let allowOpenWith = (Extension.desktopManager.getNumberOfSelectedItems() == 1);
                 this._actionOpenWith.setSensitive(allowOpenWith);
@@ -717,10 +755,13 @@ var FileItem = class {
         info.set_attribute_string('metadata::nautilus-icon-position',
                                   `${pos[0]},${pos[1]}`);
         this.file.set_attributes_async(info,
-            Gio.FileQueryInfoFlags.NONE,
-            GLib.PRIORITY_DEFAULT,
-            this._setMetadataCancellable,
-            (source, result) => this._onSetMetadataFileFinished(source, result)
+                                       Gio.FileQueryInfoFlags.NONE,
+                                       GLib.PRIORITY_DEFAULT,
+                                       this._setMetadataCancellable,
+            (source, result) => {
+                this._setMetadataCancellable = null;
+                this._onSetMetadataFileFinished(source, result);
+            }
         );
     }
 
@@ -775,7 +816,7 @@ var FileItem = class {
     }
 
     get trustedDesktopFile() {
-        return this._isDesktopFile &&
+        return this._isValidDesktopFile &&
                this._attributeCanExecute &&
                this.metadataTrusted &&
                !Extension.desktopManager.writableByOthers &&
